#!/usr/bin/env python
##
# @file       fuota.py
# @author     SIANA Systems
# @date       08/2023
# @copyright  The MIT License (MIT)
# @brief      BLE Firmware.Upgrade utility.
#
# This script is used to perform a firmware upgrade over BLE implementing ST's
# FUOTA protocol for STM32WB/WBA devices. In addition, allows to scan for
# specific devices or retrieve their bdaddr
#
# Installing dependencies:
#    $: pip install bleak
#    
# Usage: fuota.py [-b BDADDRESS ] [IMAGE.bin]
#
# Options:
#   -h, --help            show this help message and exit
#   -b, --bdaddress       target BDADDRESS (default: first come)
#   -a, --address         image offset from base on target device (default: 0x7c000)
#   -v, --verbose         Verbose mode (debug)
#
# Examples:
#
#   1. upgrade a device:
#    > fuota.py -b D8:67:D6:EA:FB:A0 firmware.bin
#
# ------------------------------------------------------------------------------
# The MIT License (MIT)
# 
# Copyright (c) 2023 Siana Systems
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#  
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

__author__ = "SIANA Systems"
__license__ = "MIT"
__version__ = "ALPHA"
__email__ = "support@siana-systems.com"

import os
import sys
import time
import asyncio
import struct
import optparse

from bleak import BleakClient

verbose = False

F150_FLASH_PAGE_SIZE = 0x100 

class Fuota:
    """Implements the client FUOTA
    """
    # FUOTA service:
    uuid_service = '0000fe20-cc7a-482a-984a-7f2ed5b3e58f'
    uuid_base_addr_char = '0000fe22-8e22-4541-9d4c-21edae82ed19'
    uuid_confirm_char = '0000fe23-8e22-4541-9d4c-21edae82ed19'
    uuid_raw_data_char = '0000fe24-8e22-4541-9d4c-21edae82ed19'

    MTU_SIZE = 251

    # STM32WB(a) Sector size
    ARCH_SECTOR_SIZE = 0x2000

    # OTA Action codes
    STOP_ALL_UPLOAD = 0x00
    USER_CONF_UPLOAD = 0x01
    APPLICATION_UPLOAD = 0x02
    END_FILE_TRANSFER = 0x06
    UPLOAD_FINISHED = 0x07
    CANCEL_UPLOAD = 0x08

    # OTA Conf codes
    REBOOT_CONFIRMED = 0x01
    READY_TO_RECEIVE_FILE = 0x02
    NOT_READY_TO_RECEIVE_FILE = 0x03

    # BFU readable action codes
    code_str = {
        STOP_ALL_UPLOAD: "STOP_ALL_UPLOAD",
        USER_CONF_UPLOAD: "USER_CONF_UPLOAD",
        APPLICATION_UPLOAD: "APPLICATION_UPLOAD",
        END_FILE_TRANSFER: "END_FILE_TRANSFER",
        UPLOAD_FINISHED: "UPLOAD_FINISHED",
        CANCEL_UPLOAD: "CANCEL_UPLOAD",
    }

    def __init__(self):
        # variables
        self.client = None
        self.file_data = None
        self.base_addr = 0
        self.image_size = 0
        self.total_sent = 0
        self.status_code = 0
        self.ota_start = 0.0
        self.ota_duration = 0.0
        self.confirm_evt = None
        self.packet_size = 0
        self.nb_chunks = 0

    async def _confirm_char_handler(self, sender: str, data):
        # We receive this indication at three points: Can't receive file, Ready To receive, and Finished/reboot
        self.status_code = data[0]
        if self.status_code == self.NOT_READY_TO_RECEIVE_FILE:
            # We notify the user that the device is not ready, and finish the script
            print("> Device: NOT Ready to receive File! - Terminating Script.")
            self.confirm_evt.set()
        if self.status_code == self.READY_TO_RECEIVE_FILE:
            # We start the File transfer process
            print("> Device: Ready to receive File! - Starting transfer.")
            self.confirm_evt.set()
        if self.status_code == self.REBOOT_CONFIRMED:
            # We are done - Notify User and finish the script
            print("> Device: Finalized Image reception - Rebooting.")
            self.confirm_evt.set()

    async def update_firmware(self, bdaddress: str, address: str, filepath: str):
        # reset for new session
        self.total_sent = 0

        print(f"> connecting to {bdaddress}...", end='')
        async with BleakClient(bdaddress) as client:
            print(" CONNECTED!")
            try:
                # adapter warm up (?)
                await asyncio.sleep(1.0)
                self.client = client

                if not client.is_connected:
                    print("> unable to connect!")
                    return False
                    
                print("Pairing...")
                client.pair()

                self.confirm_evt = asyncio.Event()

                # determine the packet DATA size
                self.packet_size = client.mtu_size - 3
                if (self.packet_size > F150_FLASH_PAGE_SIZE):
                    self.packet_size = F150_FLASH_PAGE_SIZE
                print(f" MTU is {client.mtu_size} => packet is {self.packet_size}")

                # cache image in memory...
                with open(filepath, 'rb') as file:
                    self.file_data = file.read()
                self.image_size = len(self.file_data)

                self.base_addr = int(address, 16)

                # Safety check on BaseAddr
                if self.base_addr > 0xFEFF0:
                    print(">> ERROR: Image Offset higher than max allowed!")
                    exit(-1)

                print(">> Registering for notifications")
                # register for notification...
                await client.start_notify(
                    self.uuid_confirm_char, self._confirm_char_handler
                )

                print(">> Sending APPLICATION_UPLOAD")
                await self._send_base_addr(self.client, self.APPLICATION_UPLOAD)

                # wait for confirmation message...
                self.ota_start = time.time()
                await self.confirm_evt.wait()
                self.confirm_evt.clear()

                print(">> Starting the file transfer")
                # File Transfer Loop
                while (self.status_code == self.READY_TO_RECEIVE_FILE) and (self.total_sent < self.image_size):
                    # Here we start sending the chunks
                    chunk_size = self.image_size - self.total_sent
                    if chunk_size > self.packet_size:
                        chunk_size = self.packet_size
                    await self._send_chunk(self.client, chunk_size)

                # Finished sending everything. Signal End of File and wait for the final indication
                await self._send_base_addr(self.client, self.UPLOAD_FINISHED)
                # This if is to avoid waiting if we reached here due to early termination
                if self.status_code == self.READY_TO_RECEIVE_FILE:
                    print("> Image Fully queued in Host - Waiting for BLE Transfer to finish.")
                    print("Do not close the script until the process finishes.")
                    await self.confirm_evt.wait()
                    self.ota_duration = int(time.time() - self.ota_start)
                await client.disconnect()
                print("> disconnected!")

            except Exception as exp:
                print(f"!!OTA: {exp}")

    async def _send_base_addr(self, client: BleakClient, code):
        pkt = bytearray(4)
        pkt[0] = code
        pkt[1] = self.base_addr & 0x0000FF
        pkt[2] = (self.base_addr & 0x00FF00 >> 8)
        pkt[3] = (self.base_addr & 0xF00F00 >> 16)
        #pkt[4] = int(self.image_size / self.ARCH_SECTOR_SIZE)
        #if self.image_size % self.ARCH_SECTOR_SIZE != 0:
        #    pkt[4] += 1
        await client.write_gatt_char(
            self.uuid_base_addr_char,
            pkt,
            response=False
        )

    async def _send_chunk(self, client: BleakClient, chunk_size):
        # send one "chunk" in a packet (240 bytes)
        pkt_data = self.file_data[self.total_sent: (self.total_sent + chunk_size)]
        await client.write_gatt_char(
            self.uuid_raw_data_char,
            pkt_data,
            response=False
        )
        self.total_sent += chunk_size
        # display transfer metrics
        self.nb_chunks += 1
        percent_file = int((self.total_sent / self.image_size) * 100.0)
        print(f"[{self.nb_chunks}]{self.total_sent}/{self.image_size} => {percent_file}%")

    # ------------------------------------------------------------------------------


if __name__ == "__main__":

    # create option parser
    p = optparse.OptionParser(description='FUOTA: BLE Firmware Update Over The Air v' + __version__)
    p.add_option('--verbose', '-v', action="store_true", help="Verbose mode")
    # connect options...
    connect_group = optparse.OptionGroup(p, "Connection Options")
    connect_group.add_option('--bdaddress', '-b', type="string", help="target BDADDRESS", default=None)
    connect_group.add_option('--address', '-a', type="string", help="Image offset on device", default="0x7c000")
    p.add_option_group(connect_group)
    # parse args/options...
    options, args = p.parse_args()

    print("\n** FUOTA: BLE Firmware Update Over The Air v" + __version__ + " (" + sys.platform + ") **\n")
    print("")
    print(" Usage:")
    print("\t- Ctl-C to exit")
    print("")

    # is BDADDRESS specified?
    if options.bdaddress is None:
        print(">> ERROR: you must specify a BDADDRESS!")
        exit(-1)

        # check if file exists
    image_file = args[0]
    if not os.path.exists(image_file):
        print(f">> ERROR: no such file: {image_file}")
        exit(-2)

    try:
        # create event loop
        loop = asyncio.get_event_loop()
        fuota = Fuota()

        print(f"> sending firmware: {image_file}")
        up_fw = asyncio.ensure_future(fuota.update_firmware(options.bdaddress.strip(), options.address, image_file))
        loop.run_until_complete(up_fw)

        # display summary
        if fuota.ota_duration > 0:
            ota_datarate = (fuota.total_sent / fuota.ota_duration) / 1024.0
            print(
                f"> OTA took {fuota.ota_duration} sec. to send {fuota.total_sent} bytes at {ota_datarate:.2f} kB/sec.")
        print("> done!")

    except Exception as ex:
        print(f"\n>> EXCEPTION: {ex}")
    except KeyboardInterrupt:
        print()
        print("> user stopped program.")
