#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
   Script to program new devices with all the firmware parts:
   - (bootloader + Application) FULL image
   - FUS
   - BLE stack
   - encryption key
   - Option BytesWarning
"""
# --------------------------------------
__author__ = "SIANA Systems"
__version__ = "ALPHA"
# -------------------------------------

import pathlib as pl
import subprocess as sp
import sys
import os
import typing as tp
import shutil as sh
import re
import fnmatch

# -->> Definitions <<------------------
PATH_SCRIPTS = pl.Path(__file__).parent
"""pl.Path: Script ROOT path"""
PATH_PARENT = PATH_SCRIPTS / ".."
"""pl.Path: Script parent path"""
PATH_BINARIES = PATH_SCRIPTS / "../Binaries"

stm32programmercli="STM32_Programmer_CLI"
"""Name of the executable"""

EXE_STM32CUBEPROG:  tp.Optional[pl.Path] = None
"""CubeProg executable"""

# ---------------------------------------------------------------------------------

#: STM32CubeProg executable
cubeProg:   tp.Optional[pl.Path] = None

# CubeProg Options
connect_no_reset=" -c port=SWD mode=HotPlug"
connect=" -c port=SWD mode=UR --hardRst"

# CubeProg commands
# FUS version
read_ver_cmd=" -r32 0x20030030 1"
expected_read = "0x20030030 : "

#start the FUS
startfus_cmd = " -startfus"
startfus_expected = "StartFus activated successfully"
fusgetstate_cmd = " -fusgetstate"
fus_expected = "FUS_STATE_NOT_RUNNING"

# Stack Version
read_stackVer_cmd = " -r32 0x20010014 1"
expected_resp_stackVer = "0x20010014 : "
NO_STACK_VERSION=0xFFFFFFFF

# Stack upgrade
stack_file =  str(PATH_BINARIES.absolute()) + "/stm32wb5x_BLE_Stack_full_fw.bin"
#start address for the stack may change according to release version, 
#please review Release Notes.
stack_address = 0x080CE000   # address corresponding to BLE Full Stack v 117.2
stack_cmd=" -fwupgrade " + str(stack_file) + " " + str(hex(stack_address)) + " firstinstall=0 startstack=1 -v"

#Mass erase
erase_cmd=" --erase all"
erase_expected_resp="Mass erase successfully achieved"

# Provision User key
key_file = str(PATH_BINARIES.absolute()) + "/OEM_KEY_COMPANY1_key_AES_CBC.bin"
key_cmd=" -wusrkey " + str(key_file) + " keytype=1"
key_expected_resp="Usrkey command  execution finished"

# Flash FULL image
full_image_name = "CaireDiagnostics-F150-App-FULL-*.hex"
#full_image_file = str(PATH_BINARIES.absolute()) + "/CaireDiagnostics-F150-App-FULL-v3.8.2336500455.hex"
#flash_cmd=" -w " + str(full_image_file) + " 0x08000000 -v"
flash_expected_response="Download verified successfully"

#Read protection
rdp_0_val=0xAA
rdp_1_val=0xBB
rdp_2_val=0xCC
OB_expected_result="Option Bytes successfully programmed"
OB_expected_result_locked="Error: Unable to reconnect after setting the Option Bytes" 

# ---------------------------------------------------------------------------------

# F150 device configuration:

MIN_FUS_VERSION=0x01020000

DEVICE_STACK_VERSION=0x01110200

# RDP level to be applied:
rdp_selected_val=rdp_2_val

# WRP starts at 0x08000000 to 0x08017000
wrp_conf=" -ob WRP1A_STRT=0x00 -ob WRP1A_END=0x17"

# ---------------------------------------------------------------------------------

def which(
    exe:        str,
    required:   bool = True
) -> tp.Optional[pl.Path]:
    """
    Get the path to the provided command.

    :param str exe:         Command or commands (separated by or) to find
    :param bool required:   Force find
    """
    match = [item for item in [sh.which(cmd=item.strip()) for item in exe.split("|")] if item is not None]
    if required and not match:
        raise FileNotFoundError(f"Unable to find {exe} executable on PATH")
    return pl.Path(match.pop(0)) if match else None


def get_exe_stm32cubeProgCLI() -> pl.Path:
    """
    Retrieves the STM32CubeProg CLI executable.

    :return: Executable path
    :rtype: pl.Path
    """
    global EXE_STM32CUBEPROG
    if not EXE_STM32CUBEPROG:
        EXE_STM32CUBEPROG = which(exe="STM32_Programmer_CLI | STM32_Programmer_cli")
    return EXE_STM32CUBEPROG

def get_FUS_version():
    """
    Read the FUS version from the device

    :return: FUS version
    :rtype: int
    """
    fus_ver = -1
    src=cubeProg.parent
    run_cmd = stm32programmercli + connect_no_reset + read_ver_cmd
    print("Reading FUS version:\n")
    print(">> ",run_cmd)
    result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)
    
    match = re.search(re.escape(expected_read), str(result.stdout))
    if match:
        # Start position of the substring in the long string
        start_pos = match.end()
        # Extract the following characters
        fus_ver = int( str(result.stdout)[start_pos:start_pos + 8], 16)
        
        if(fus_ver == 0):  #failed to read
            #start FUS
            run_cmd = stm32programmercli + connect + startfus_cmd
            print("FUS is not running, start FUS:\n")
            print(">> ",run_cmd)
            result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)
            
            match = re.search(re.escape(startfus_expected), str(result.stdout))
            if match:

                run_cmd = stm32programmercli + connect_no_reset + read_ver_cmd
                print("Retrying to read FUS version:\n")
                print(">> ",run_cmd)
                result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)
              
                match = re.search(re.escape(expected_read), str(result.stdout))
                if match:
                    # Start position of the substring in the long string
                    start_pos = match.end()
                    # Extract the following characters
                    fus_ver = int( str(result.stdout)[start_pos:start_pos + 8], 16)     
            else:
                print("  Failed to start FUS\n\n")
                print(result.stdout)
                exit(-1)

    else:
        print("  Failed to read the fUS version\n\n")
        print(result.stdout)
        exit(-1)
 
    return fus_ver
   
def get_Stack_version():
    """
    Read the BLE stack version from the device

    :return: BLE stack version
    :rtype: int
    """
    stack_ver = -1
    src=cubeProg.parent
    
    run_cmd = stm32programmercli + connect_no_reset + read_stackVer_cmd
    print("Reading the current stack version")
    print(">> ",run_cmd)
    result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)
 
    match = re.search(re.escape(expected_resp_stackVer), str(result.stdout))
    if match:
        # Start position of the substring in the long string
        start_pos = match.end()
        # Extract the following characters
        stack_ver = int( str(result.stdout)[start_pos:start_pos + 8], 16)
    else:
        print("  Failed to read the fUS version\n\n")
        print(result.stdout)
        exit(-1)
 
    return stack_ver
   
   
def upgrade_FUS():
    """
    Upgrade the FUS version

    :return: None
    """
    print("Upgrading FUS           - not yet implemented")
    exit(-1)
    
    
def erase_all(): 
    """
    Mass erase the device

    :return: None
    """
    src=cubeProg.parent
    run_cmd = stm32programmercli + connect_no_reset + erase_cmd
    print("Erasing all the flash:\n")
    print(">> ",run_cmd)
    result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)
    #print(result.returncode)
    #print(result.stderr)
    #print(result.stdout)
    
    match = re.search(re.escape(erase_expected_resp), str(result.stdout))
    if (result.returncode == 0) and match:
        print("  Mass erase SUCCESS\n")
    else:
        print("  Failed to MASS Erase the device\n\n")
        print(result.stdout)
        exit(-1)

    
def upgrade_STACK():
    """
    Upgrade the BLE stack

    :return: None
    """
    print("Upgrading STACK - please wait...")
    src=cubeProg.parent
    
    #getstate x 3
    run_cmd = stm32programmercli + connect_no_reset + fusgetstate_cmd
    result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)
    result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)
    result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)

    run_cmd = stm32programmercli + connect_no_reset + stack_cmd
    print(">> ",run_cmd) 
    result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)
    
    #wait until done:
    done = False
    run_cmd = stm32programmercli + connect_no_reset + fusgetstate_cmd
    while (not done):
        result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)
        match = re.search(re.escape(fus_expected), str(result.stdout))
        if match:
            done = True
    
    
def provision_UserKey():
    """
    Provision the User key

    :return: None
    """
    src=cubeProg.parent
    run_cmd = stm32programmercli + connect_no_reset + key_cmd
    print("Provisioning User key:\n")
    print(key_file)
    print(">> ",run_cmd)
    result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)
    
    match = re.search(re.escape(key_expected_resp), str(result.stdout))
    if (result.returncode == 0) and match:
        print("  User Key Provision SUCCESS\n")
    else:
        print("  Failed to flash the User Key\n\n")
        print(result.stdout)
        exit(-1)

    
def flash_FULL_binary():
    """
    Flash the FULL firmware image
    
    :return: None
    """
    print("Flashing FULL image:\n")

    full_image_file = [n for n in fnmatch.filter(os.listdir(PATH_BINARIES), full_image_name) ]
    if (len(full_image_file) == 1):
        full_image_file = str(PATH_BINARIES.absolute()) + "/" + full_image_file[0]
        flash_cmd=" -w " + str(full_image_file) + " 0x08000000 -v"
        
        src=cubeProg.parent
        run_cmd = stm32programmercli + connect_no_reset + flash_cmd
        print(">> ",run_cmd)
        result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)

        match = re.search(re.escape(flash_expected_response), str(result.stdout))
        if (result.returncode == 0) and match:
            print("  Flash full image SUCCESS\n")
        else:
            print("  Failed to flash the full image\n\n")
            print(result.stdout)
            exit(-1)
    else:
        print("  Failed to find the FULL image to flash!")
        exit(-1)

def update_OptionBytes():
    """
    Update the Option Bytes (read and write protections)

    :return: None
    """
    print("Updating Option Bytes (read and write protections)")

    src=cubeProg.parent
    print("Setting WRP for sectors 0 to 17" )
    run_cmd = stm32programmercli + connect_no_reset + wrp_conf
    print(">> ",run_cmd)
    result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)

    # Check if stdout contains OB_expected_result
    match = re.search(re.escape(OB_expected_result), str(result.stdout))
    if (result.returncode == 0) and match:
        print("  WRP setting SUCCESS\n")
    else:
        print("  Failed to set WRP\n\n")
        print(result.stdout)
        exit(-1)
    
    print("Setting RDP level to ", rdp_selected_val )
    rdp_conf=" -ob RDP="+str(rdp_selected_val)
    run_cmd = stm32programmercli + connect_no_reset + rdp_conf
    print(">> ",run_cmd)
    result = sp.run(run_cmd, cwd=src, capture_output=True, shell=True)

    # Check if stdout contains OB_expected_result
    match = re.search(re.escape(OB_expected_result), str(result.stdout))
    match_failed_after = re.search(re.escape(OB_expected_result_locked), str(result.stdout))
    if match or match_failed_after:
        print("  RDP level SUCCESS\n")
    else:
        print("  Failed to set RDP level\n\n")
        print(result.stdout)
        exit(-1)

    
if __name__ == "__main__":

    print("\nF150 Production Programmer v" + __version__ + " (" + sys.platform + ") **\n")
    print("")
    
    cubeProg = get_exe_stm32cubeProgCLI()
    
    # Get the current FUS version
    current_fus_version = get_FUS_version()
    print("  Current FUS version is ", hex(current_fus_version))
     
    # check version, need to upgrade FUS?
    if (current_fus_version < MIN_FUS_VERSION):
        print("  Need to upgrade FUS")
        upgrade_FUS()
        
    # Mass erase
    erase_all()
    
    #Check BLE stack version:
    current_stack = get_Stack_version()
    print("  Current Stack version is ", hex(current_stack))
     
    # check version, need to upgrade FUS?
    if (current_stack < DEVICE_STACK_VERSION) and (current_stack != NO_STACK_VERSION):
        print("  Need to upgrade the BLE Stack")
        # upgrade the BLE stack
        upgrade_STACK()
    
    # Crypto Key
    provision_UserKey()
    
    # Flash the main firmware
    flash_FULL_binary()
    
    
    # Update Option Bytes (read and write protections)
    update_OptionBytes()
    
    print("All done!!")
    exit(0)
    
